import argparse
import datetime
import math
import typing as t

import randog.factory
from ..._utils.type import timedelta, positive_timedelta
from ... import timedelta_util
from .. import Args, Subcmd
from ._base import SubcmdDef, add_common_arguments


class SubcmdDefTimedelta(SubcmdDef):
    def cmd(self) -> Subcmd:
        return Subcmd.Timedelta

    def add_parser(self, subparsers) -> argparse.ArgumentParser:
        timedelta_parser = subparsers.add_parser(
            Subcmd.Timedelta.value,
            usage=(
                "python -m randog timedelta [MINIMUM MAXIMUM] [--unit UNIT] "
                "[--iso | --fmt FORMAT] [common-options]"
            ),
            description=(
                "It generates values of type datetime.timedelta. "
                "This mode uses the simple format to represent timedelta; "
                "such as '30d', '1h30m' and '1d20h30m40s'"
            ),
            add_help=False,
        )
        timedelta_args_group = timedelta_parser.add_argument_group("arguments")
        timedelta_args_group.add_argument(
            "minimum",
            type=timedelta,
            nargs="?",
            metavar="MINIMUM",
            help=(
                "the minimum value with the simple format such as '1d20h30m40s'. "
                "If not specified, the behavior is left to the specification of "
                "randog.factory.randtimedelta."
            ),
        )
        timedelta_args_group.add_argument(
            "maximum",
            type=timedelta,
            nargs="?",
            metavar="MAXIMUM",
            help=(
                "the maximum value with the simple format such as '1d20h30m40s'. "
                "If not specified, the behavior is left to the specification of "
                "randog.factory.randtimedelta."
            ),
        )
        timedelta_args_group.add_argument(
            "--unit",
            type=positive_timedelta,
            metavar="UNIT",
            help=(
                "the minimum unit of generated values. "
                "If not specified, the behavior is left to the specification of "
                "randog.factory.randtimedelta."
            ),
        )
        group_date_fmt = timedelta_args_group.add_mutually_exclusive_group()
        group_date_fmt.add_argument(
            "--iso",
            action="store_true",
            help="if specified, it outputs generated object with ISO-8601 format",
        )
        group_date_fmt.add_argument(
            "--fmt",
            dest="format",
            metavar="FORMAT",
            help="if specified, it outputs generated object with the specified format, "
            "such as '%%tH:%%M:%%S.%%f'",
        )
        add_common_arguments(timedelta_parser)

        return timedelta_parser

    def validate_parser(self, args: Args, subparser: argparse.ArgumentParser):
        if args.sub_cmd != Subcmd.Timedelta:
            return

        iargs, kwargs = self.build_args(args)
        minimum, maximum = iargs
        unit = kwargs["unit"]

        if minimum is not None and maximum is not None and minimum > maximum:
            subparser.error("arguments must satisfy MINIMUM <= MAXIMUM")

        if None not in (unit, minimum, maximum):
            min_by_unit = math.ceil(minimum / unit)
            max_by_unit = math.floor(maximum / unit)
            if min_by_unit > max_by_unit:
                subparser.error(
                    "argument --unit: "
                    "there is no multiple of the unit value between MINIMUM and MAXIMUM"
                )

        if args.output_fmt == "repr" and args.iso:
            subparser.error("argument --iso: not allowed with argument --repr")
        elif args.output_fmt == "repr" and args.format:
            subparser.error("argument --fmt: not allowed with argument --repr")

    def build_args(
        self, args: Args
    ) -> t.Tuple[t.Sequence[t.Any], t.Mapping[str, t.Any]]:
        return (args.get("minimum"), args.get("maximum")), {"unit": args.get("unit")}

    def get_factory_constructor(self) -> t.Callable:
        return lambda *a, **kw: randog.factory.randtimedelta(*a, **kw).post_process(
            _TimedeltaWrapper
        )


class _TimedeltaWrapper(datetime.timedelta):
    _base: datetime.timedelta

    def __new__(cls, base: datetime.timedelta):
        ins = super(_TimedeltaWrapper, cls).__new__(
            cls,
            days=base.days,
            seconds=base.seconds,
            microseconds=base.microseconds,
        )
        ins._base = base
        return ins

    def __str__(self):
        return timedelta_util.to_str(self)

    def __repr__(self):
        return repr(self._base)

    def __format__(self, format_spec):
        return timedelta_util.to_fmt(self, format_spec)

    def isoformat(self):
        return timedelta_util.to_iso(self)



from enum import Enum


class Subcmd(Enum):
    Byfile = "byfile"
    Bool = "bool"
    Int = "int"
    Float = "float"
    String = "str"
    Decimal = "decimal"
    Datetime = "datetime"
    Date = "date"
    Time = "time"
    Timedelta = "timedelta"



from ._base import Factory, REGENERATE_PROB_MAX, FactoryStopException
from ._choice import ChoiceRandomFactory, randchoice, randenum
from ._const import const
from ._bool import BoolRandomFactory, randbool
from ._int import IntRandomFactory, randint
from ._float import FloatRandomFactory, randfloat
from ._decimal import DecimalRandomFactory, randdecimal
from ._str import StrRandomFactory, randstr
from ._time import TimeRandomFactory, randtime
from ._date import DateRandomFactory, randdate
from ._timedelta import TimedeltaRandomFactory, randtimedelta
from ._datetime import DatetimeRandomFactory, randdatetime
from ._list import randlist
from ._dict import DictItem, DictRandomFactory, randdict
from ._by_callable import ByCallableFactory, by_callable
from ._by_iterator import ByIteratorFactory, by_iterator
from ._increment import increment
from ._union import UnionRandomFactory, union
from ._from_example import from_example, FromExampleContext
from ._from_pyfile import from_pyfile, FactoryDef

__all__ = [
    "from_example",
    "FromExampleContext",
    "from_pyfile",
    "Factory",
    "randchoice",
    "randenum",
    "const",
    "randbool",
    "randint",
    "randfloat",
    "randdecimal",
    "randstr",
    "randtime",
    "randdate",
    "randtimedelta",
    "randdatetime",
    "randlist",
    "randdict",
    "by_callable",
    "by_iterator",
    "increment",
    "union",
    "DictItem",
    "FactoryDef",
    "FactoryStopException",
    "REGENERATE_PROB_MAX",
]



import argparse
import typing as t

import randog.factory
from ..._utils.type import probability
from .. import Args, Subcmd
from ._base import SubcmdDef, add_common_arguments


class SubcmdDefFloat(SubcmdDef):
    def cmd(self) -> Subcmd:
        return Subcmd.Float

    def add_parser(self, subparsers) -> argparse.ArgumentParser:
        float_parser = subparsers.add_parser(
            Subcmd.Float.value,
            usage=(
                "python -m randog float [MINIMUM MAXIMUM] [--p-inf PROB_P_INF] "
                "[--n-inf PROB_N_INF] [--nan PROB_NAN] [common-options]"
            ),
            description="It generates values of type float.",
            add_help=False,
        )
        float_args_group = float_parser.add_argument_group("arguments")
        float_args_group.add_argument(
            "minimum",
            type=float,
            nargs="?",
            metavar="MINIMUM",
            help=(
                "the minimum value. "
                "If not specified, the behavior is left to the specification of "
                "randog.factory.randfloat."
            ),
        )
        float_args_group.add_argument(
            "maximum",
            type=float,
            nargs="?",
            metavar="MAXIMUM",
            help=(
                "the maximum value. "
                "If not specified, the behavior is left to the specification of "
                "randog.factory.randfloat."
            ),
        )
        float_args_group.add_argument(
            "--p-inf",
            type=probability,
            default=0.0,
            metavar="PROB_P_INF",
            help="the probability of positive infinity; default=0.0",
        )
        float_args_group.add_argument(
            "--n-inf",
            type=probability,
            default=0.0,
            metavar="PROB_N_INF",
            help="the probability of negative infinity; default=0.0",
        )
        float_args_group.add_argument(
            "--nan",
            type=probability,
            default=0.0,
            metavar="PROB_NAN",
            help="the probability of NaN; default=0.0",
        )
        add_common_arguments(float_parser)

        return float_parser

    def validate_parser(self, args: Args, subparser: argparse.ArgumentParser):
        if args.sub_cmd != Subcmd.Float:
            return

        iargs, kwargs = self.build_args(args)
        minimum, maximum = iargs
        nan = kwargs["nan"]
        p_inf = kwargs["p_inf"]
        n_inf = kwargs["n_inf"]

        if minimum is not None and maximum is not None and minimum > maximum:
            subparser.error("arguments must satisfy MINIMUM <= MAXIMUM")

        if nan + p_inf + n_inf > 1.0:
            subparser.error(
                "arguments must satisfy that PROB_P_INF + PROB_N_INF + PROB_NAN <= 1.0"
            )

    def build_args(
        self, args: Args
    ) -> t.Tuple[t.Sequence[t.Any], t.Mapping[str, t.Any]]:
        return (args.get("minimum"), args.get("maximum")), {
            "p_inf": args.get("p_inf"),
            "n_inf": args.get("n_inf"),
            "nan": args.get("nan"),
        }

    def get_factory_constructor(self) -> t.Callable:
        return randog.factory.randfloat



import typing as t

from ._base import Factory

T = t.TypeVar("T")


def by_iterator(iterator: t.Iterator[T]) -> Factory[T]:
    """Return a factory generating values by specified iterator.

    Parameters
    ----------
    iterator : Iterator[T]
        the iterator generating value
    """
    return ByIteratorFactory(iterator)


class ByIteratorFactory(Factory[T]):
    """factory generating values by specified iterator"""

    _iterator: t.Iterator[T]

    def __init__(
        self,
        iterator: t.Iterator[T],
    ):
        """Return a factory generating values by specified iterator.

        Parameters
        ----------
        iterator : Iterator[T]
            the iterator generating value
        """
        self._iterator = iterator

    def _next(self) -> T:
        return next(self._iterator)



from ._examples import DictItemExample, Example
from . import factory
from ._main import RandogCmdWarning

__all__ = [
    "factory",
    "DictItemExample",
    "Example",
    "RandogCmdWarning",
]

__version__ = "0.11.0"



import dataclasses
from os import PathLike
import types
import typing as t

from randog.factory import Factory


@dataclasses.dataclass
class FactoryDef:
    factory: Factory
    csv_columns: t.Optional[t.Sequence[str]]


@t.overload
def from_pyfile(
    file: t.Union[str, PathLike, t.IO], *, full_response: t.Literal[False] = False
) -> Factory:
    pass


@t.overload
def from_pyfile(
    file: t.Union[str, PathLike, t.IO], *, full_response: t.Literal[True]
) -> FactoryDef:
    pass


def from_pyfile(
    file: t.Union[str, PathLike, t.IO],
    *,
    full_response: bool = False,
) -> t.Union[FactoryDef, Factory]:
    """Returns a factory defined in the specified file.

    Parameters
    ----------
    file : str | PathLike | IO
        the filename of the factory definition
    full_response : bool (default=False)
        If True is specified, the return value is the FactoryDef dataclass,
        and data other than the factory can be obtained.
    """
    if isinstance(file, (str, PathLike)):
        with open(file, mode="rb") as fp:
            factory_def = _from_pyfile(fp, file)
    else:
        factory_def = _from_pyfile(file, "<io>")

    if full_response:
        return factory_def
    else:
        return factory_def.factory


FACTORY_ATTR_NAME = "FACTORY"
CSV_COL_ATTR_NAME = "CSV_COLUMNS"


def _from_pyfile(fp: t.IO, filename: t.Union[str, PathLike]) -> FactoryDef:
    import randog

    d = types.ModuleType("factory")
    d.randog = randog
    d.__file__ = str(filename)
    try:
        exec(compile(fp.read(), filename, "exec"), d.__dict__)
    except OSError as e:
        e.strerror = f"Unable to load factory file ({e.strerror})"
        raise

    return _load_factory_module(d, filename=filename)


def _load_factory_module(
    module: types.ModuleType, *, filename: t.Union[str, PathLike]
) -> FactoryDef:
    if not hasattr(module, FACTORY_ATTR_NAME):
        raise AttributeError(
            f"factory file '{filename}' has no attribute '{FACTORY_ATTR_NAME}'"
        )

    factory = getattr(module, FACTORY_ATTR_NAME)
    if not isinstance(factory, Factory):
        raise AttributeError(
            f"attribute '{FACTORY_ATTR_NAME}' "
            f"of factory file '{filename}' is not a factory"
        )

    csv_columns = getattr(module, CSV_COL_ATTR_NAME, None)
    if csv_columns is not None and not isinstance(csv_columns, t.Sequence):
        raise AttributeError(
            f"attribute '{CSV_COL_ATTR_NAME}' "
            f"of factory file '{filename}' MUST be None or a sequence of strings"
        )

    return FactoryDef(
        factory=factory,
        csv_columns=csv_columns,
    )



import csv
import datetime
import json
import os
import random
import sys
import typing as t
import warnings

import randog.factory
from ._logging import (
    logger,
    apply_stderr_logging_config,
    apply_logging_config_file,
    apply_default_logging_config,
)
from ._warning import RandogCmdWarning, apply_formatwarning
from .._utils.exceptions import get_message_recursive
from ..factory import FactoryDef, FactoryStopException
from . import Args, Subcmd, get_subcmd_def


def _build_factories(
    args: Args,
) -> t.Iterator[t.Tuple[int, str, randog.factory.Factory]]:
    subcmd_def = get_subcmd_def(args.sub_cmd)

    if args.sub_cmd == Subcmd.Byfile:
        for factory_count, filepath in enumerate(args.factories):
            if filepath == "-":
                def_file_name = ""
                factory_def = randog.factory.from_pyfile(sys.stdin, full_response=True)
            else:
                def_file_name = os.path.basename(filepath)
                if def_file_name.endswith(".py"):
                    def_file_name = def_file_name[:-3]
                factory_def = randog.factory.from_pyfile(filepath, full_response=True)
            factory = factory_def.factory

            post_process = _gen_post_function_for_byfile_mode(args, factory_def)
            if post_process is not None:
                factory = factory.post_process(post_process)

            yield factory_count, def_file_name, factory
    else:
        iargs, kwargs = subcmd_def.build_args(args)
        factory = subcmd_def.get_factory_constructor()(*iargs, **kwargs)
        if args.iso:
            factory = factory.post_process(
                lambda x: x.isoformat() if x is not None else None
            )
        elif args.format:
            factory = factory.post_process(
                lambda x: format(x, args.format) if x is not None else None
            )
        yield 0, "", factory


def _get_csv_field(pre_value: t.Mapping, col) -> t.Any:
    if isinstance(col, t.Callable):
        return col(pre_value)
    else:
        return pre_value.get(col)


def _gen_post_function_for_byfile_mode(
    args: Args, factory_def: FactoryDef
) -> t.Optional[t.Callable[[t.Any], t.Any]]:
    """args に応じて、factory_def.factory に適用する post_process 関数を作成する。

    この関数は、byfile モードでのみ使用する。
    """
    if args.sub_cmd is not Subcmd.Byfile:
        raise RuntimeError("internal error; Unexpected function call")

    # CSV 出力の場合、CSVの行 (Sequence[Any]) を生成するような post_process を作成する。
    if args.get("csv", False):
        if factory_def.csv_columns is not None:

            def _post_function(pre_value) -> t.Optional[t.Sequence[t.Any]]:
                if pre_value is None:
                    return None
                elif isinstance(pre_value, t.Mapping):
                    return [
                        _get_csv_field(pre_value, col)
                        for col in factory_def.csv_columns
                    ]
                elif isinstance(pre_value, t.Sequence) and not isinstance(
                    pre_value, str
                ):
                    return pre_value
                else:
                    warnings.warn(
                        "--csv is recommended for only collections (such as "
                        "dict, list, tuple, etc.); "
                        "In CSV output, one generated value is treated as one row, "
                        "so the result is the same as --repeat except for collections; "
                        "CSV_COLUMNS in the definition file is ignored.",
                        RandogCmdWarning,
                    )
                    return [pre_value]

        else:

            def _post_function(pre_value) -> t.Optional[t.Sequence[t.Any]]:
                if pre_value is None:
                    return None
                elif isinstance(pre_value, t.Mapping):
                    # Why msg_part is defined:
                    #   Consider the possibility that future modifications will allow
                    #   users to reach here in other than byfile mode,
                    #   and branch the error message.
                    msg_part = (
                        " in the definition file"
                        if args.sub_cmd is Subcmd.Byfile
                        else ""
                    )
                    warnings.warn(
                        f"Since CSV_COLUMNS is not defined{msg_part}, "
                        "the fields are inserted in the order returned by the "
                        "dictionary; In this case, fields may not be aligned "
                        "depending on the FACTORY definition, "
                        "so it is recommended to define CSV_COLUMNS.",
                        RandogCmdWarning,
                    )
                    return list(pre_value.values())
                elif isinstance(pre_value, t.Sequence) and not isinstance(
                    pre_value, str
                ):
                    return pre_value
                else:
                    warnings.warn(
                        "--csv is recommended for only collections (such as "
                        "dict, list, tuple, etc.); "
                        "In CSV output, one generated value is treated as one row, "
                        "so the result is the same as --repeat except for collections.",
                        RandogCmdWarning,
                    )
                    return [pre_value]

        return _post_function
    else:
        return None


class _DummyIO:
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


def _open_output_fp(
    args: Args,
    number: int,
    *,
    def_file: str,
    repeat_count: int,
    factory_count: int,
    now: datetime.datetime,
    already_written_files: t.MutableSet[str],
) -> t.Union[_DummyIO, t.TextIO]:
    output_path = args.output_path_for(
        number,
        def_file=def_file,
        repeat_count=repeat_count,
        factory_count=factory_count,
        now=now,
        env=os.environ,
    )

    if output_path is None:
        return _DummyIO()
    else:
        options = {}

        if output_path in already_written_files:
            options["mode"] = "at"
        else:
            options["mode"] = "wt"
            already_written_files.add(output_path)

        if args.get("csv", False):
            options["newline"] = ""
        elif args.output_linesep is not None:
            options["newline"] = args.output_linesep

        if args.output_encoding is not None:
            options["encoding"] = args.output_encoding

        logger.debug(f"open file: {output_path}")

        return open(output_path, **options)


def _output_generated(generated: t.Any, fp: t.TextIO, args: Args):
    if args.output_fmt == "repr":
        print(repr(generated), file=fp)
    elif args.output_fmt == "json":
        json.dump(generated, fp, default=str)
        fp.write("\n")
    else:
        print(generated, file=fp)


def _output_to_csv(
    factory: randog.factory.Factory[t.Optional[t.Sequence[t.Sequence[t.Any]]]],
    line_num: int,
    fp: t.TextIO,
    regenerate: float,
    discard: float,
    raise_on_factory_stopped: bool,
    linesep: t.Optional[str],
):
    writer_options = {}
    if fp in (sys.stdout, sys.stderr):
        # windows で "\r\n" にすると、標準出力時に改行が "\r\r\n" に変換されてしまう。
        # よって、windows で改行を "\r\n" にしたい場合も、標準出力時はここには　"\n" を指定する。
        writer_options["lineterminator"] = "\n"
    elif linesep is None:
        writer_options["lineterminator"] = os.linesep
    else:
        writer_options["lineterminator"] = linesep

    csv_writer = csv.writer(fp, **writer_options)
    csv_writer.writerows(
        filter(
            lambda x: x is not None,
            factory.iter(
                line_num,
                regenerate=regenerate,
                discard=discard,
                raise_on_factory_stopped=raise_on_factory_stopped,
            ),
        )
    )


class _Discarded(Exception):
    pass


def _generate_according_args(
    args: Args,
    factory: randog.factory.Factory,
) -> t.Tuple[t.Any, t.Literal["generated", "stop_iter", "discarded"]]:
    """generated a value according arguments

    This method generates values assuming it is not a csv output.

    Parameters
    ----------
    args
        the arguments
    factory
        the factory

    Returns
    -------
    Tuple[Any, Literal["generated", "stop_iter", "discarded"]]
        the generated value and a return type

    Raises
    ------
    RuntimeError
        When CSV output is specified.
        This method generates values assuming it is not a csv output.
    """
    if args.csv:
        raise RuntimeError("cannot use this method if --csv is specified")

    if args.list is not None:
        generated = list(
            factory.iter(
                args.list,
                regenerate=args.regenerate,
                discard=args.discard,
                raise_on_factory_stopped=args.error_on_factory_stopped,
            )
        )
    else:
        try:
            while True:
                generated = factory.next(
                    raise_on_factory_stopped=args.error_on_factory_stopped
                )
                if random.random() >= args.regenerate:
                    break
        except StopIteration:
            return None, "stop_iter"
        if random.random() < args.discard:
            return None, "discarded"

    return generated, "generated"


def _setup_primary_configuration(args: Args):
    """Setup primary python configuration

    It performs a setup that should be performed as soon as the arguments are
    determined in accordance with the python specification.

    This method should be executed once and only once when randog is run as command.

    Parameters
    ----------
    args:
        arguments of command execution
    """

    # setup logging
    if args.log_stderr:
        apply_stderr_logging_config(args.log_stderr)
    elif args.log_config_file:
        try:
            apply_logging_config_file(args.log_config_file)
        except Exception as e:
            logger.error(
                "failed to apply the logging configure file; "
                f"{'; '.join(get_message_recursive(e))}",
            )
            # If yaml is missed, warn about it, as that might cause Exception
            try:
                import yaml

                assert yaml is not None
            except ModuleNotFoundError:
                logger.error(
                    "Are you trying to use YAML format logging configuration? "
                    "If you want to use YAML format configuration files, "
                    "PyYAML must be installed."
                )
            exit(1)
    else:
        apply_default_logging_config()

    # setup warning
    if args.hide_randog_warnings:
        warnings.simplefilter("ignore", RandogCmdWarning)


def main():
    apply_formatwarning()

    args = Args(sys.argv)
    now = datetime.datetime.now()
    already_written_files = set()

    _setup_primary_configuration(args)

    # set environments
    os.environ.update(args.env)

    logger.info(f"run randog with args: {args.commanded_args}")

    try:
        index = 0
        for factory_count, def_file, factory in _build_factories(args):
            for r_index in range(args.repeat):
                with _open_output_fp(
                    args,
                    index,
                    def_file=def_file,
                    repeat_count=r_index,
                    factory_count=factory_count,
                    now=now,
                    already_written_files=already_written_files,
                ) as fp_opened:
                    index += 1
                    # args に応じて出力先 fp を決定する。
                    if not isinstance(fp_opened, _DummyIO):
                        fp = fp_opened
                    else:
                        fp = sys.stdout

                    # 生成処理と出力処理
                    # CSV 出力の場合に生成の方法が異なるので、生成と出力をひとまとめにした。
                    if args.csv is not None:
                        _output_to_csv(
                            factory,
                            args.csv,
                            fp,
                            regenerate=args.regenerate,
                            discard=args.discard,
                            raise_on_factory_stopped=args.error_on_factory_stopped,
                            linesep=args.output_linesep,
                        )
                        logger.debug(
                            "output to CSV; "
                            f"factory_count={factory_count}, repeat_count={r_index}"
                        )
                    else:
                        generated, gen_result = _generate_according_args(args, factory)

                        if gen_result == "stop_iter":
                            break
                        elif gen_result == "discarded":
                            continue
                        else:
                            _output_generated(generated, fp, args=args)
                            logger.debug(
                                "output generated; "
                                f"factory_count={factory_count}, repeat_count={r_index}"
                            )
    except FactoryStopException:
        logger.error("the factory stopped generating before the process was complete")
        exit(1)
    except Exception as e:
        logger.error("; ".join(get_message_recursive(e)), exc_info=e)
        exit(1)



import logging


def get_logger(name: str = "") -> logging.Logger:
    if name is None or len(name) == 0:
        logger = logging.getLogger("randog")
    else:
        logger = logging.getLogger(f"randog.{name}")

    logger.addHandler(logging.NullHandler())
    return logger



import copy
import json
import logging
import logging.config
import os
import typing as t

from .._logging import get_logger as _get_logger

logger = _get_logger("cmd")


class CmdLogFormatter(logging.Formatter):
    def format(self, record):
        custom_record = copy.copy(record)

        custom_record.levelname = record.levelname.lower()
        custom_record.exc_info = None
        return logging.Formatter.format(self, custom_record)


def apply_default_logging_config():
    root_logger = logging.getLogger()
    root_logger.setLevel("WARNING")
    root_logger.addHandler(logging.NullHandler())

    handler = logging.StreamHandler()
    handler.setLevel("ERROR")
    handler.formatter = CmdLogFormatter("%(levelname)s: %(message)s")
    logger.addHandler(handler)


def apply_stderr_logging_config(level: str):
    root_logger = logging.getLogger()
    root_logger.setLevel(level)
    handler = logging.StreamHandler()
    handler.formatter = CmdLogFormatter("%(levelname)s: %(message)s")
    root_logger.addHandler(handler)


def apply_logging_config_file(config_file: t.Union[str, os.PathLike]):
    # まずエラー出力を設定する。
    # dictConfig 適用中のエラーも出力できるようにするために、この段階で実施する。
    handler = logging.StreamHandler()
    handler.setLevel("ERROR")
    handler.formatter = CmdLogFormatter("%(levelname)s: %(message)s")
    logger.addHandler(handler)

    # apply the logging configuration file
    try:
        import yaml

        with open(config_file, mode="rt") as fp:
            config = yaml.load(fp, yaml.Loader)
    except ModuleNotFoundError:
        with open(config_file, mode="rt") as fp:
            config = json.load(fp)
    logging.config.dictConfig(config)

    # dictConfig によってエラー出力が上書きされた場合に備えて、再度エラー出力を設定する。
    # ただし、明示的に disable_existing_loggers=True が指定されている場合は上書きがユーザの意思であると判断して、再設定はしない。
    if not config.get("disable_existing_loggers", False):
        logger.removeHandler(handler)
        logger.addHandler(handler)



import string
import typing as t
from random import Random

from .._base import Factory
from ..._utils.nullsafe import dfor
from ...exceptions import FactoryConstructionError


@t.overload
def randstr(
    *,
    length: t.Union[int, Factory[int], None] = None,
    charset: t.Optional[str] = None,
    regex: None = None,
    rnd: t.Optional[Random] = None,
) -> Factory[str]:
    pass


@t.overload
def randstr(
    *,
    length: None = None,
    charset: None = None,
    regex: t.Union[str, t.Pattern],
    rnd: t.Optional[Random] = None,
) -> Factory[str]:
    pass


def randstr(
    *,
    length: t.Union[int, Factory[int], None] = None,
    charset: t.Optional[str] = None,
    regex: t.Optional[str] = None,
    rnd: t.Optional[Random] = None,
) -> Factory[str]:
    """Return a factory generating random str values.

    Parameters
    ----------
    length : int|Factory[int], default=8
        length of generated string
    charset : str, optional
        characters to be used
    regex : str, optional
        regular expression for generated string.
        It cannot be used with `length` or `charset`.
    rnd : Random, optional
        random number generator to be used

    Raises
    ------
    FactoryConstructionError
        When the specified generating conditions are inconsistent.
    """
    if regex is not None and (length is not None or charset is not None):
        raise FactoryConstructionError(
            "cannot specify argument 'regex' for randstr() with 'length' or 'charset'"
        )

    if regex is not None:
        try:
            from ._regex import StrRegexRandomFactory
        except ImportError:
            raise FactoryConstructionError(
                "package 'rstr' must be installed to specify 'regex' to randstr()"
            )
        return StrRegexRandomFactory(regex=regex, rnd=rnd)
    else:
        return StrRandomFactory(length=length, charset=charset, rnd=rnd)


class StrRandomFactory(Factory[str]):
    """factory generating random int values"""

    _random: Random
    _length: t.Union[int, Factory[int]]
    _charset: str

    def __init__(
        self,
        *,
        length: t.Union[int, Factory[int], None] = None,
        charset: t.Optional[str] = None,
        rnd: t.Optional[Random] = None,
    ):
        """Return a factory generating random str values.

        Parameters
        ----------
        length : int|Factory[int], default=8
            length of generated string
        charset : str, optional
            characters to be used
        rnd : Random, optional
            random number generator to be used

        Raises
        ------
        FactoryConstructionError
            When the specified generating conditions are inconsistent.
        """
        self._random = dfor(rnd, Random())
        self._length = dfor(length, 8)
        self._charset = dfor(charset, string.ascii_letters + string.digits)

        if isinstance(self._length, Factory) and len(self._charset) == 0:
            raise FactoryConstructionError(
                "the charset for randstr() must not be empty if length is at random"
            )
        if (
            not isinstance(self._length, Factory)
            and self._length > 0
            and len(self._charset) == 0
        ):
            raise FactoryConstructionError(
                "the charset for randstr() must not be empty if length is positive"
            )

    def _next(self) -> str:
        length = self._next_length()
        return "".join(self._random.choices(self._charset, k=length))

    def _next_length(self) -> int:
        if isinstance(self._length, Factory):
            return self._length.next()
        else:
            return self._length



import typing as t
from decimal import Decimal
from random import Random

from ._base import Factory
from ._float import randfloat
from .._utils.nullsafe import dfor, dforc


def randdecimal(
    minimum: t.Optional[t.SupportsFloat] = None,
    maximum: t.Optional[t.SupportsFloat] = None,
    *,
    decimal_len: t.Optional[int] = None,
    p_inf: t.SupportsFloat = 0.0,
    n_inf: t.SupportsFloat = 0.0,
    nan: t.SupportsFloat = 0.0,
    rnd: t.Optional[Random] = None,
) -> Factory[Decimal]:
    """Return a factory generating random Decimal values.

    Parameters
    ----------
    minimum : float, optional
        the minimum
    maximum : float, optional
        the maximum
    decimal_len : int, optional
        the length of decimal part
    p_inf : float, default=0
        the probability of positive infinity
    n_inf : float, default=0
        the probability of negative infinity
    nan : float, default=0
        the probability of NaN
    rnd : Random, optional
        random number generator to be used

    Raises
    ------
    FactoryConstructionError
        When the specified generating conditions are inconsistent.
    """
    return DecimalRandomFactory(
        minimum,
        maximum,
        decimal_len=decimal_len,
        p_inf=p_inf,
        n_inf=n_inf,
        nan=nan,
        rnd=rnd,
    )


class DecimalRandomFactory(Factory[Decimal]):
    """factory generating random Decimal values"""

    _random: Random
    _factory: Factory[float]
    _decimal_len: t.Optional[Decimal]

    def __init__(
        self,
        minimum: t.Optional[t.SupportsFloat] = None,
        maximum: t.Optional[t.SupportsFloat] = None,
        *,
        decimal_len: t.Optional[int] = None,
        p_inf: t.SupportsFloat = 0.0,
        n_inf: t.SupportsFloat = 0.0,
        nan: t.SupportsFloat = 0.0,
        rnd: t.Optional[Random] = None,
    ):
        """Return a factory generating random Decimal values.

        Parameters
        ----------
        minimum : float, optional
            the minimum
        maximum : float, optional
            the maximum
        decimal_len : int, optional
            the length of decimal part
        p_inf : float, default=0
            the probability of positive infinity
        n_inf : float, default=0
            the probability of negative infinity
        nan : float, default=0
            the probability of NaN
        rnd : Random, optional
            random number generator to be used

        Raises
        ------
        FactoryConstructionError
            When the specified generating conditions are inconsistent.
        """
        self._random = dfor(rnd, Random())
        self._factory = randfloat(
            minimum, maximum, p_inf=p_inf, n_inf=n_inf, nan=nan, rnd=self._random
        )
        self._decimal_len = dforc(lambda x: Decimal(1).scaleb(-x), decimal_len)

    def _next(self) -> Decimal:
        pre_value = self._factory.next()
        value = Decimal(pre_value)

        if self._decimal_len is not None and value.is_finite():
            value = value.quantize(self._decimal_len)

        return value



import datetime as dt
import math
import typing as t
from random import Random

from ._base import Factory
from ._int import randint
from .._utils.nullsafe import dfor
from ..exceptions import FactoryConstructionError


_ZERO = dt.timedelta(0)
_DAY = dt.timedelta(days=1)
_HOUR = dt.timedelta(hours=1)
_MINUTE = dt.timedelta(minutes=1)
_SECOND = dt.timedelta(seconds=1)
_MILLISECOND = dt.timedelta(milliseconds=1)
_MICROSECOND = dt.timedelta(microseconds=1)


_NEXT = {
    _MICROSECOND: _MILLISECOND,
    _MILLISECOND: _SECOND,
    _SECOND: _MINUTE,
    _MINUTE: _HOUR,
    _HOUR: _DAY,
    _DAY: 30 * _DAY,
}


def randtimedelta(
    minimum: t.Optional[dt.timedelta] = None,
    maximum: t.Optional[dt.timedelta] = None,
    *,
    unit: t.Optional[dt.timedelta] = None,
    rnd: t.Optional[Random] = None,
) -> Factory[dt.timedelta]:
    """Return a factory generating random timedelta values.

    Parameters
    ----------
    minimum : timedelta, optional
        the minimum
    maximum : timedelta, optional
        the maximum
    unit : timedelta, optional
        the atomic unit
    rnd : Random, optional
        random number generator to be used

    Raises
    ------
    FactoryConstructionError
        When the specified generating conditions are inconsistent.
    """
    return TimedeltaRandomFactory(
        minimum,
        maximum,
        unit=unit,
        rnd=rnd,
    )


class TimedeltaRandomFactory(Factory[dt.timedelta]):
    """factory generating random timedelta values"""

    _random: Random
    _inner_factory: Factory[int]
    _min: dt.timedelta
    _max: dt.timedelta
    _min_by_unit: int
    _max_by_unit: int
    _unit: dt.timedelta

    def __init__(
        self,
        minimum: t.Optional[dt.timedelta] = None,
        maximum: t.Optional[dt.timedelta] = None,
        *,
        unit: t.Optional[dt.timedelta] = None,
        rnd: t.Optional[Random] = None,
    ):
        """Return a factory generating random timedelta values.

        Parameters
        ----------
        minimum : timedelta, optional
            the minimum
        maximum : timedelta, optional
            the maximum
        unit : timedelta, optional
            the atomic unit
        rnd : Random, optional
            random number generator to be used

        Raises
        ------
        FactoryConstructionError
            When the specified generating conditions are inconsistent.
        """
        if unit is not None:
            if unit == dt.timedelta(0):
                raise FactoryConstructionError(
                    "the unit for randtimedelta must not be zero"
                )
            if unit < dt.timedelta(0):
                unit = -unit

        self._random = dfor(rnd, Random())
        self._min, self._max, self._unit = self._normalize(minimum, maximum, unit)

        if self._min > self._max:
            raise FactoryConstructionError("empty range for randtimedelta")

        self._min_by_unit = math.ceil(self._min / self._unit)
        self._max_by_unit = math.floor(self._max / self._unit)
        if self._min_by_unit > self._max_by_unit:
            raise FactoryConstructionError("empty range for randtimedelta")
        self._inner_factory = randint(self._min_by_unit, self._max_by_unit, rnd=rnd)

    def _next(self) -> dt.timedelta:
        return self._unit * self._inner_factory.next()

    @classmethod
    def _normalize(
        cls,
        minimum: t.Optional[dt.timedelta] = None,
        maximum: t.Optional[dt.timedelta] = None,
        unit: t.Optional[dt.timedelta] = None,
    ) -> t.Tuple[dt.timedelta, dt.timedelta, dt.timedelta]:
        if minimum is not None and maximum is not None:
            if unit is None:
                unit = calc_unit(minimum, maximum)
            return minimum, maximum, unit

        elif minimum is not None:
            if unit is None:
                unit = calc_unit(minimum, maximum)
                return minimum, minimum + _NEXT[unit], unit
            else:
                return minimum, minimum + 10 * unit, unit

        elif maximum is not None:
            if unit is None:
                unit = calc_unit(minimum, maximum)
                return maximum - _NEXT[unit], maximum, unit
            else:
                return maximum - 10 * unit, maximum, unit

        else:
            if unit is None:
                return _ZERO, _DAY, _HOUR
            else:
                return _ZERO, _NEXT[unit], unit


def calc_unit(
    *timedelta_list: t.Optional[dt.timedelta], default: dt.timedelta = _DAY
) -> dt.timedelta:
    result = default
    for timedelta in filter(lambda x: x is not None, timedelta_list):
        if timedelta.microseconds != 0:
            if result > _MICROSECOND and timedelta.microseconds % 1000 != 0:
                result = _MICROSECOND
            elif result > _MILLISECOND:
                result = _MILLISECOND
        elif result <= _MILLISECOND:
            continue
        elif timedelta.seconds != 0:
            if result > _SECOND and timedelta.seconds % 60 != 0:
                result = _SECOND
            elif result > _MINUTE and timedelta.seconds % 3600 != 0:
                result = _MINUTE
            elif result > _HOUR:
                result = _HOUR

    return result



import datetime as dt
import typing as t
from random import Random

from ._base import Factory
from .._utils.nullsafe import dfor
from ..exceptions import FactoryConstructionError


def randdatetime(
    minimum: t.Union[dt.datetime, dt.date, None] = None,
    maximum: t.Union[dt.datetime, dt.date, None] = None,
    *,
    tzinfo: t.Union[t.Literal[False], None, dt.tzinfo] = False,
    rnd: t.Optional[Random] = None,
) -> Factory[dt.datetime]:
    """Return a factory generating random datetime values.

    Parameters
    ----------
    minimum : datetime | date, optional
        the minimum
    maximum : datetime | date, optional
        the maximum
    tzinfo : tzinfo | None, optional
        If specified, the tzinfo of result will be fixed to this value
        (False means no specification).
        When it fixes aware datetime to aware, the time is corrected.
        Otherwise, only the tzinfo is changed.
    rnd : Random, optional
        random number generator to be used

    Raises
    ------
    FactoryConstructionError
        When the specified generating conditions are inconsistent.
    """
    return DatetimeRandomFactory(
        minimum,
        maximum,
        fix_timezone=tzinfo is not False,
        fixed_timezone=tzinfo if tzinfo is not False else None,
        rnd=rnd,
    )


class DatetimeRandomFactory(Factory[dt.datetime]):
    """factory generating random datetime values"""

    _random: Random
    _min: dt.datetime
    _max: dt.datetime
    _range: dt.timedelta
    _fix_timezone: bool
    _fixed_timezone: t.Optional[dt.tzinfo]

    def __init__(
        self,
        minimum: t.Union[dt.datetime, dt.date, None] = None,
        maximum: t.Union[dt.datetime, dt.date, None] = None,
        *,
        fix_timezone: bool = False,
        fixed_timezone: t.Optional[dt.tzinfo] = None,
        rnd: t.Optional[Random] = None,
    ):
        """Return a factory generating random datetime values.

        Parameters
        ----------
        minimum : datetime | date, optional
            the minimum
        maximum : datetime | date, optional
            the maximum
        fix_timezone : bool, default=False
            If it is True, the tzinfo of result is fixed to `fixed_timezone`.
        fixed_timezone : tzinfo, default=None
            If `fix_timezone` is True, the tzinfo of result is fixed to this.
        rnd : Random, optional
            random number generator to be used

        Raises
        ------
        FactoryConstructionError
            When the specified generating conditions are inconsistent.
        """
        self._random = dfor(rnd, Random())
        self._min, self._max = self._normalize(minimum, maximum)
        self._fix_timezone = fix_timezone
        self._fixed_timezone = fixed_timezone

        if (self._min.tzinfo is None and self._max.tzinfo is not None) or (
            self._min.tzinfo is not None and self._max.tzinfo is None
        ):
            raise FactoryConstructionError(
                "cannot define range for randdatetime with a naive datetime "
                "and an aware datetime"
            )
        if self._min > self._max:
            raise FactoryConstructionError("empty range for randdatetime")

        self._range = self._max - self._min

    def _next(self) -> dt.datetime:
        weight = self._random.random()

        pre_result = self._min + self._range * weight
        if self._fix_timezone:
            if pre_result.tzinfo is not None and self._fixed_timezone is not None:
                return pre_result.astimezone(self._fixed_timezone)
            else:
                return pre_result.replace(tzinfo=self._fixed_timezone)
        else:
            return pre_result

    @classmethod
    def _normalize(
        cls,
        minimum: t.Union[dt.datetime, dt.date, None],
        maximum: t.Union[dt.datetime, dt.date, None],
    ) -> t.Tuple[dt.datetime, dt.datetime]:
        if isinstance(minimum, dt.date) and not isinstance(minimum, dt.datetime):
            minimum = dt.datetime.combine(minimum, dt.time(0))
        if isinstance(maximum, dt.date) and not isinstance(maximum, dt.datetime):
            maximum = dt.datetime.combine(
                maximum + dt.timedelta(days=1), dt.time(0)
            ) - dt.timedelta(microseconds=1)

        if minimum is not None and maximum is not None:
            return minimum, maximum
        elif minimum is not None:
            return minimum, minimum + dt.timedelta(days=1)
        elif maximum is not None:
            return maximum - dt.timedelta(days=1), maximum
        else:
            now = dt.datetime.utcnow()
            return now - dt.timedelta(hours=12), now + dt.timedelta(hours=12)



import codecs
import datetime as dt
from .. import timedelta_util


def positive_int(value):
    value_int = int(value)
    if value_int <= 0:
        raise ValueError("must be positive")
    return value_int


def non_negative_int(value):
    value_int = int(value)
    if value_int < 0:
        raise ValueError("must be non-negative")
    return value_int


def probability(value):
    value_float = float(value)
    if 0 <= value_float <= 1:
        return value_float
    else:
        raise ValueError("must be in the range [0, 1]")


def datetime(value):
    return dt.datetime.fromisoformat(value)


def time(value):
    return dt.time.fromisoformat(value)


def date(value):
    return dt.date.fromisoformat(value)


def timedelta(value):
    return timedelta_util.from_str(value)


def positive_timedelta(value):
    result = timedelta_util.from_str(value)
    if result <= dt.timedelta(0):
        raise ValueError("must be positive")
    return timedelta_util.from_str(value)


def encoding(value):
    try:
        codecs.lookup(value)
    except LookupError as e:
        raise ValueError(*e.args)
    return value



import typing as t

from ._base import Factory

T = t.TypeVar("T")


def by_callable(func: t.Callable[[], T]) -> Factory[T]:
    """Return a factory generating values by specified callable.

    Parameters
    ----------
    func : () -> T
        the function generating value
    """
    return ByCallableFactory(func)


class ByCallableFactory(Factory[T]):
    """factory generating values by specified callable"""

    _callable: t.Callable[[], T]

    def __init__(
        self,
        func: t.Callable[[], T],
    ):
        """Return a factory generating values by specified callable.

        Parameters
        ----------
        func : () -> T
            the function generating value
        """
        self._callable = func

    def _next(self) -> T:
        return self._callable()



from random import Random
import typing as t

from ._base import Factory
from .._utils.nullsafe import dfor
from ..exceptions import FactoryConstructionError


def randbool(
    prop_true: float = 0.5,
    *,
    rnd: t.Optional[Random] = None,
) -> Factory[bool]:
    """Return a factory generating random bool values.

    Parameters
    ----------
    prop_true : float
        the probability of True
    rnd : Random, optional
        random number generator to be used

    Raises
    ------
    FactoryConstructionError
        When the specified generating conditions are inconsistent.
    """
    return BoolRandomFactory(prop_true, rnd=rnd)


class BoolRandomFactory(Factory[bool]):
    """factory generating random bool values"""

    _random: Random
    _prop_true: float

    def __init__(
        self,
        prop_true: float,
        *,
        rnd: t.Optional[Random] = None,
    ):
        """Return a factory generating random bool values.

        Parameters
        ----------
        prop_true : float
            the probability of True
        rnd : Random, optional
            random number generator to be used

        Raises
        ------
        FactoryConstructionError
            When the specified generating conditions are inconsistent.
        """
        self._random = dfor(rnd, Random())
        self._prop_true = prop_true

        if self._prop_true < 0.0 or 1.0 < self._prop_true:
            raise FactoryConstructionError(
                "the probability `prob_true` must range from 0 to 1"
            )

    def _next(self) -> int:
        return self._random.random() < self._prop_true



import argparse
from abc import ABC, abstractmethod
import typing as t

from .. import Linesep
from ..._utils.type import positive_int, encoding
from .._subcmd import Subcmd
from .._main_args import Args


class SubcmdDef(ABC):
    @abstractmethod
    def cmd(self) -> Subcmd:
        ...

    @abstractmethod
    def add_parser(self, subparsers) -> argparse.ArgumentParser:
        ...

    @abstractmethod
    def validate_parser(self, args: Args, subparser: argparse.ArgumentParser):
        ...

    @abstractmethod
    def build_args(
        self, args: Args
    ) -> t.Tuple[t.Sequence[t.Any], t.Mapping[str, t.Any]]:
        ...

    @abstractmethod
    def get_factory_constructor(self) -> t.Callable:
        ...


def add_common_arguments(parser: argparse.ArgumentParser):
    common_opt_group = parser.add_argument_group("common-options")
    other_opt_group = parser.add_argument_group("other options")
    group_output_fmt = common_opt_group.add_mutually_exclusive_group()
    common_opt_group.add_argument(
        "--repeat",
        "-r",
        metavar="COUNT",
        default=1,
        type=positive_int,
        help=(
            "repeat generation a specified number of times. "
            "The results are output one by one; if you want them as a single list, "
            "use --list instead."
        ),
    )
    common_opt_group.add_argument(
        "--list",
        "-L",
        metavar="LENGTH",
        type=positive_int,
        help=(
            "if specified, repeats the specified numerical generation "
            "and returns a list consisting of the results."
        ),
    )
    group_output_fmt.add_argument(
        "--repr",
        dest="output_fmt",
        action="store_const",
        const="repr",
        help="if specified, it outputs generated object by repr()",
    )
    group_output_fmt.add_argument(
        "--json",
        dest="output_fmt",
        action="store_const",
        const="json",
        help="if specified, it outputs generated object by json format",
    )
    common_opt_group.add_argument(
        "--output",
        "-O",
        metavar="DESC_PATH",
        help="destination file path",
    )
    common_opt_group.add_argument(
        "--output-encoding",
        "-X",
        metavar="ENCODING",
        default=None,
        type=encoding,
        help="encoding for output. "
        "This is only effective when outputting to a file with the '--output'/'-O' "
        "option.",
    )
    common_opt_group.add_argument(
        "--output-linesep",
        "--O-ls",
        default=None,
        choices=Linesep.names(),
        help="line separator for output. "
        "This is only effective when outputting to a file with the '--output'/'-O' "
        "option.",
    )
    common_opt_group.add_argument(
        "--quiet",
        "-q",
        action="store_true",
        help=(
            "hide warnings of randog. "
            "If you want to hide all warnings, use -W option of python."
        ),
    )
    group_logging = common_opt_group.add_mutually_exclusive_group()
    group_logging.add_argument(
        "--log-stderr",
        default=None,
        choices=("ERROR", "WARNING", "INFO", "DEBUG"),
        help="output logs of specified level or more stronger into standard error.",
    )
    group_logging.add_argument(
        "--log",
        metavar="LOGGING_CONFIG_PATH",
        default=None,
        help="logging configuration file (JSON or YAML)",
    )
    common_opt_group.add_argument(
        "--env",
        metavar="VAR=VAL",
        action="append",
        nargs="+",
        help="additional environment variable",
    )
    other_opt_group.add_argument(
        "-h", "--help", action="help", help="show this help message and exit"
    )



import typing as t


def get_message_recursive(e: Exception) -> t.List[str]:
    result = []

    head_e = e
    for _ in range(32):  # against infinite loops
        result.append(f"{head_e.__class__.__name__}: {head_e}")
        if head_e.__cause__ is not None:
            head_e = head_e.__cause__
        else:
            break

    return result



import warnings


def apply_formatwarning():
    origin_formatwarning = warnings.formatwarning

    def _formatwarning(message, category, *args, **kwargs):
        if issubclass(category, RandogCmdWarning):
            if isinstance(message, Warning):
                return f"warning: {message.args[0]}\n"
            else:
                return f"warning: {message}\n"
        else:
            return origin_formatwarning(message, category, *args, **kwargs)

    warnings.formatwarning = _formatwarning


class RandogCmdWarning(FutureWarning):
    """warning raised during command execution"""

    pass



from ._warning import RandogCmdWarning
from ._linesep import Linesep
from ._subcmd import Subcmd
from ._main_args import Args
from ._subcmd_def import iter_subcmd_def, get_subcmd_def
from ._main_func import main



import argparse
import typing as t

import randog.factory
from .. import Args, Subcmd
from ._base import SubcmdDef, add_common_arguments


class SubcmdDefInt(SubcmdDef):
    def cmd(self) -> Subcmd:
        return Subcmd.Int

    def add_parser(self, subparsers) -> argparse.ArgumentParser:
        int_parser = subparsers.add_parser(
            Subcmd.Int.value,
            usage="python -m randog int MINIMUM MAXIMUM [common-options]",
            description="It generates integer values.",
            add_help=False,
        )
        int_args_group = int_parser.add_argument_group("arguments")
        int_args_group.add_argument(
            "minimum",
            type=int,
            metavar="MINIMUM",
            help="the minimum value",
        )
        int_args_group.add_argument(
            "maximum",
            type=int,
            metavar="MAXIMUM",
            help="the maximum value",
        )
        add_common_arguments(int_parser)

        return int_parser

    def validate_parser(self, args: Args, subparser: argparse.ArgumentParser):
        if args.sub_cmd != Subcmd.Int:
            return

        iargs, kwargs = self.build_args(args)
        minimum, maximum = iargs

        if minimum is not None and maximum is not None and minimum > maximum:
            subparser.error("arguments must satisfy MINIMUM <= MAXIMUM")

    def build_args(
        self, args: Args
    ) -> t.Tuple[t.Sequence[t.Any], t.Mapping[str, t.Any]]:
        return (args.get("minimum"), args.get("maximum")), {}

    def get_factory_constructor(self) -> t.Callable:
        return randog.factory.randint



from dataclasses import dataclass
from random import Random
import typing as t

from ._base import Factory
from .._utils.nullsafe import dfor
from ..exceptions import FactoryConstructionError


_item_tuple = t.Tuple[Factory, float]


@dataclass
class DictItem:
    """
    A rule for generating values corresponding to a key in the random generation of
    dict.
    """

    factory: Factory
    """a factory to generate values"""

    prop_exists: float
    """probability of key generation"""

    @t.overload
    def __init__(self, factory: Factory):
        ...

    @t.overload
    def __init__(self, factory: Factory, prop_exists: float):
        ...

    @t.overload
    def __init__(self, item: _item_tuple):
        ...

    def __init__(self, *args):
        if len(args) == 1 and isinstance(args[0], Factory):
            self.prop_exists = 1.0
            self.factory = args[0]
        elif len(args) == 1 and isinstance(args[0], tuple):
            self.prop_exists = args[0][1]
            self.factory = args[0][0]
        elif len(args) == 2:
            self.factory = args[0]
            self.prop_exists = args[1]
        else:
            raise DictItemValueError()


class DictItemValueError(ValueError):
    pass


def randdict(
    items_dict: t.Optional[
        t.Mapping[t.Hashable, t.Union[Factory, _item_tuple, DictItem]]
    ] = None,
    *,
    rnd: t.Optional[Random] = None,
    **items: t.Union[Factory, _item_tuple, DictItem],
) -> Factory[dict]:
    """Return a factory generating random dict.

    Parameters
    ----------
    items : Mapping
        the factories of each key.
        If `items_dict` is specified, `items` will be ignored.
    items_dict: Mapping
        the factories of each key. Use when keyword arguments cannot be specified.
    rnd : Random, optional
        random number generator to be used
    """
    if items_dict is not None:
        items = items_dict

    items_normalized = {}
    non_factory_item_keys = []
    for key, value in items.items():
        try:
            items_normalized[key] = (
                value if isinstance(value, DictItem) else DictItem(value)
            )
        except DictItemValueError:
            non_factory_item_keys.append(str(key))
    if len(non_factory_item_keys) > 0:
        raise FactoryConstructionError(
            "randdict received non-factory object for item: "
            f"{', '.join(non_factory_item_keys)}"
        )

    return DictRandomFactory(items_normalized, rnd=rnd)


class DictRandomFactory(Factory[dict]):
    """factory generating random dict"""

    _random: Random
    _items: t.Mapping[t.Hashable, DictItem]

    def __init__(
        self,
        items: t.Mapping[t.Hashable, DictItem],
        *,
        rnd: t.Optional[Random] = None,
    ):
        """Return a factory generating random dict.

        Parameters
        ----------
        items : Mapping
            the factories of each key
        rnd : Random, optional
            random number generator to be used
        """
        self._random = dfor(rnd, Random())
        self._items = items

    def _next(self) -> dict:
        result = {}

        for key, cond in self._items.items():
            if self._random.random() < cond.prop_exists:
                result[key] = cond.factory.next()

        return result



import argparse
import typing as t

import randog.factory
from ..._utils.type import non_negative_int
from .. import Args, Subcmd
from ._base import SubcmdDef, add_common_arguments


class SubcmdDefString(SubcmdDef):
    def cmd(self) -> Subcmd:
        return Subcmd.String

    def add_parser(self, subparsers) -> argparse.ArgumentParser:
        str_parser = subparsers.add_parser(
            Subcmd.String.value,
            usage=(
                "python -m randog str [--length LENGTH] [--charset CHARSET] "
                "[--regex REGEX] [common-options]"
            ),
            description="It generates values of type str.",
            add_help=False,
        )
        str_args_group = str_parser.add_argument_group("arguments")
        str_args_group.add_argument(
            "--length",
            type=non_negative_int_range,
            default=None,
            metavar="LENGTH",
            help=(
                "the length of generated strings. "
                "You can specify an integer such as '--length 5' or a range such as "
                "'--length 3:8'."
            ),
        )
        str_args_group.add_argument(
            "--charset",
            type=str,
            default=None,
            metavar="CHARSET",
            help="the characters which contained by generated strings",
        )
        str_args_group.add_argument(
            "--regex",
            type=str,
            default=None,
            metavar="REGEX",
            help=(
                "the regular expression for generated string. "
                "It cannot be used with `--length` or `--charset`."
            ),
        )
        add_common_arguments(str_parser)

        return str_parser

    def validate_parser(self, args: Args, subparser: argparse.ArgumentParser):
        iargs, kwargs = self.build_args(args)
        regex = kwargs["regex"]
        length = kwargs.get("length")
        charset = kwargs["charset"]

        if regex is not None:
            try:
                import rstr  # noqa: F401
            except ImportError:
                subparser.error(
                    "argument --regex: package 'rstr' is required to use --regex"
                )

        if regex is not None and length is not None:
            subparser.error("argument --regex: not allowed with argument --length")
        if regex is not None and charset is not None:
            subparser.error("argument --regex: not allowed with argument --charset")

    def build_args(
        self, args: Args
    ) -> t.Tuple[t.Sequence[t.Any], t.Mapping[str, t.Any]]:
        kwargs = {"charset": args.get("charset"), "regex": args.get("regex")}
        length = args.get("length")
        if length is not None:
            if length[0] == length[1]:
                kwargs["length"] = length[0]
            else:
                kwargs["length"] = randog.factory.randint(*length)

        return tuple(), kwargs

    def get_factory_constructor(self) -> t.Callable:
        return randog.factory.randstr


def non_negative_int_range(value: str) -> t.Tuple[int, int]:
    if ":" in value:
        value_s_str, value_e_str = value.split(":", 1)
        value_s = non_negative_int(value_s_str)
        value_e = non_negative_int(value_e_str)

        if value_s > value_e:
            raise ValueError("range must not be empty")

    else:
        value_s = value_e = non_negative_int(value)

    return value_s, value_e



import datetime as dt
import enum
from decimal import Decimal
import typing as t
from random import Random

from . import (
    by_callable,
    by_iterator,
    DictItem,
    Factory,
    randbool,
    randenum,
    randtimedelta,
    randdatetime,
    randdate,
    randtime,
    randdecimal,
    randdict,
    randfloat,
    randlist,
    randint,
    randstr,
    union,
    const,
)
from ._timedelta import calc_unit
from ._logging import logger
from ..exceptions import FactoryConstructionError

import randog


def _from_decimal(
    example: Decimal,
    rnd: t.Optional[Random],
) -> Factory[Decimal]:
    p_inf, n_inf, nan = 0.0, 0.0, 0.0
    decimal_len = None
    if example.is_infinite():
        if example > 0:
            p_inf = 1.0
        else:
            n_inf = 1.0
    elif example.is_nan():
        nan = 1.0
    else:
        decimal_len = -example.as_tuple()[2]
    return randdecimal(
        p_inf=p_inf, n_inf=n_inf, nan=nan, decimal_len=decimal_len, rnd=rnd
    )


def _from_timedelta(
    example: dt.timedelta,
    rnd: t.Optional[Random],
) -> Factory[dt.timedelta]:
    unit = calc_unit(example)

    if example < dt.timedelta(0):
        return randtimedelta(None, dt.timedelta(0), unit=unit, rnd=rnd)
    else:
        return randtimedelta(dt.timedelta(0), None, unit=unit, rnd=rnd)


def _from_datetime(
    example: dt.datetime,
    rnd: t.Optional[Random],
) -> Factory[dt.datetime]:
    return randdatetime(tzinfo=example.tzinfo, rnd=rnd)


def _from_time(
    example: dt.time,
    rnd: t.Optional[Random],
) -> Factory[dt.time]:
    return randtime(tzinfo=example.tzinfo, rnd=rnd)


_FACTORY_CONSTRUCTOR_BY_TYPE: t.Dict[
    type, t.Callable[[t.Optional[Random]], Factory]
] = {
    bool: lambda rnd: randbool(rnd=rnd),
    int: lambda rnd: randint(0, 100, rnd=rnd),
    float: lambda rnd: randfloat(0, 1.0, rnd=rnd),
    Decimal: lambda rnd: randdecimal(0, 1.0, rnd=rnd),
    str: lambda rnd: randstr(rnd=rnd),
    dt.timedelta: lambda rnd: randtimedelta(rnd=rnd),
    dt.datetime: lambda rnd: randdatetime(rnd=rnd),
    dt.date: lambda rnd: randdate(rnd=rnd),
    dt.time: lambda rnd: randtime(rnd=rnd),
    list: lambda rnd: randlist(randstr(), rnd=rnd),
    tuple: lambda rnd: randlist(randstr(), type=tuple, rnd=rnd),
    dict: lambda rnd: randdict({"key": randint(0, 100, rnd=rnd)}, rnd=rnd),
}

_FACTORY_CONSTRUCTOR_BY_EXAMPLE: t.Dict[
    type, t.Callable[[t.Any, t.Optional[Random]], Factory]
] = {
    Decimal: _from_decimal,
    dt.timedelta: _from_timedelta,
    dt.datetime: _from_datetime,
    dt.time: _from_time,
}


class FromExampleContext:
    _custom_chain_length: int
    _warned_too_long_custom_chain: bool
    _path: t.Tuple[t.Any, ...]
    _custom_funcs: t.Sequence["_CustomFunc"]
    _rnd: t.Optional[Random]
    _examples_stack: t.Tuple[t.Any, ...]

    _signal_terminate_custom: bool = False

    def __init__(
        self,
        path: t.Tuple[t.Any, ...],
        custom_func: t.Union["_CustomFunc", t.Sequence["_CustomFunc"], None],
        rnd: t.Optional[Random],
        examples_stack: t.Tuple[t.Any, ...],
        custom_chain_length: int = 0,
        warned_too_long_custom_chain: bool = False,
    ):
        self._path = path
        if custom_func is None:
            self._custom_funcs = []
        elif isinstance(custom_func, t.Sequence):
            self._custom_funcs = custom_func
        else:
            self._custom_funcs = [custom_func]
        self._rnd = rnd
        self._example_stacks = examples_stack
        self._custom_chain_length = custom_chain_length
        self._warned_too_long_custom_chain = warned_too_long_custom_chain

    @property
    def custom_chain_length(self) -> int:
        return self._custom_chain_length

    @property
    def path(self) -> t.Tuple[t.Any, ...]:
        return self._path

    @property
    def rnd(self) -> t.Optional[Random]:
        return self._rnd

    @property
    def custom_funcs(self) -> t.Sequence["_CustomFunc"]:
        return self._custom_funcs

    @property
    def examples(self) -> t.Tuple[t.Any, ...]:
        return self._example_stacks

    @property
    def current_example(self) -> t.Any:
        if len(self._example_stacks) <= 0:
            return None
        else:
            return self._example_stacks[-1]

    def terminate_custom_chain(self):
        self._signal_terminate_custom = True

    @property
    def signal_terminate_custom(self) -> bool:
        return self._signal_terminate_custom

    def set_warned_too_long_custom_chain(self):
        self._warned_too_long_custom_chain = True

    @property
    def warned_too_long_custom_chain(self) -> bool:
        return self._warned_too_long_custom_chain

    def from_example(self, example: t.Any) -> Factory:
        return from_example(
            example,
            custom_func=self._custom_funcs,
            rnd=self._rnd,
            context=self,
        )

    def recursive(self, child: t.Any, key: t.Any) -> Factory:
        new_context = ContextFactory.child_of(
            self,
            key=key,
            example=child,
        )
        return from_example(
            child,
            custom_func=self._custom_funcs,
            rnd=self._rnd,
            context=new_context,
        )


class _CustomFunc(t.Protocol):
    def __call__(
        self,
        example: t.Any,
        *,
        context: FromExampleContext,
    ) -> t.Any:
        ...


class ContextFactory:
    @classmethod
    def root(
        cls,
        custom_func: t.Union[_CustomFunc, t.Sequence[_CustomFunc], None],
        rnd: t.Optional[Random],
        example: t.Any,
    ) -> FromExampleContext:
        return FromExampleContext(
            path=tuple(),
            custom_func=custom_func,
            rnd=rnd,
            examples_stack=(example,),
        )

    @classmethod
    def child_of(
        cls,
        context: FromExampleContext,
        *,
        key: t.Any,
        example: t.Any,
    ) -> FromExampleContext:
        return FromExampleContext(
            path=(*context.path, key),
            custom_func=context.custom_funcs,
            rnd=context.rnd,
            examples_stack=(*context.examples, example),
        )

    @classmethod
    def count_up_custom(
        cls,
        context: FromExampleContext,
    ) -> "FromExampleContext":
        return FromExampleContext(
            path=context.path,
            custom_func=context.custom_funcs,
            rnd=context.rnd,
            custom_chain_length=context.custom_chain_length + 1,
            warned_too_long_custom_chain=False,
            examples_stack=context.examples,
        )

    @classmethod
    def reset_signals(
        cls,
        context: FromExampleContext,
    ) -> "FromExampleContext":
        return FromExampleContext(
            path=context.path,
            custom_func=context.custom_funcs,
            rnd=context.rnd,
            custom_chain_length=context.custom_chain_length,
            warned_too_long_custom_chain=context.warned_too_long_custom_chain,
            examples_stack=context.examples,
        )


def _dict_item(obj, key, context: FromExampleContext) -> DictItem:
    if isinstance(obj, randog.DictItemExample):
        return DictItem(context.recursive(obj.example, key), obj.prop_exists)
    else:
        return DictItem(context.recursive(obj, key))


def _list_item(item: t.Tuple[int, t.Any], context: FromExampleContext) -> Factory:
    index, obj = item
    return context.recursive(obj, index)


def from_example(
    example: t.Any,
    *,
    custom_func: t.Union[_CustomFunc, t.Sequence[_CustomFunc], None] = None,
    rnd: t.Optional[Random] = None,
    context: t.Optional[FromExampleContext] = None,
) -> Factory:
    """Returns a factory generating value like specified example or type.

    Parameters
    ----------
    example : Any
        the type or the example
    custom_func : Callable | Sequence[Callable]
        If specified, this function is executed first and
        its return value is used as a new example.
        If it returns a factory, it is used as is.
        If it returns `NotImplemented`, `from_example` behaves
        as if `custom_func` was not specified.
        The context is passed to this function.
        Multiple functions may be specified for `custom_func`,
        and if multiple functions are specified,
        they are executed in sequence until a value
        other than NotImplemented is returned.
        This sequence of processing is also used to create factories
        for child elements of dict and list.
        It is recommended that `custom_func` receives `**kwargs`
        to allow for more keyword arguments in future updates.
    rnd : Random, optional
        random number generator to be used
    context : FromExampleContext, optional
        the context of generation. Normally, you should not specify it.
        If specified, the context property takes precedence over other arguments.

    Raises
    ------
    FactoryConstructionError
        When the specified example or type is not supported.
    """
    if isinstance(example, Factory):
        return example
    if context is None:
        context = ContextFactory.root(
            custom_func=custom_func,
            rnd=rnd,
            example=example,
        )

    # print(example, context.custom_chain_length, "before while")
    while True:
        # print(example, context.custom_chain_length, "while")

        # terminate custom chain if custom_func execs terminate_custom_chain()
        if context.signal_terminate_custom:
            break

        # Limit the number of customizations to avoid infinite loops
        if context.custom_chain_length >= 32:
            if not context.warned_too_long_custom_chain:
                logger.warning(
                    "The application of custom_func in this generation was stopped "
                    "because custom_func was executed more than the specified number "
                    "of times during the generation of a single value in from_example."
                )
                context.set_warned_too_long_custom_chain()
            break

        context = ContextFactory.count_up_custom(context)

        for custom_func in context.custom_funcs:
            # print(example, context.custom_chain_length, "for")
            custom_result = custom_func(
                example,
                context=context,
            )
            # print(example, custom_result, "for")
            if isinstance(custom_result, Factory):
                return custom_result
            if custom_result is not NotImplemented and example != custom_result:
                example = custom_result
                break

            # reset signals from custom_func if the example is not customized
            context = ContextFactory.reset_signals(context)
        else:
            # If customization by custom_funcs is not performed,
            # no further customization is chained.
            break

    if isinstance(example, randog.Example):
        return union(*map(context.from_example, example), rnd=context.rnd)
    if isinstance(example, type):
        if issubclass(example, enum.Enum):
            return randenum(example, rnd=context.rnd)
        elif example in _FACTORY_CONSTRUCTOR_BY_TYPE:
            return _FACTORY_CONSTRUCTOR_BY_TYPE[example](context.rnd)
        else:
            raise FactoryConstructionError(
                f"cannot construct factory for unsupported type: {example}"
            )

    if example is None:
        return const(None, rnd=context.rnd)
    for type_, construct in _FACTORY_CONSTRUCTOR_BY_EXAMPLE.items():
        if isinstance(example, type_):
            return construct(example, rnd)
    if isinstance(example, t.Mapping):
        return randdict({k: _dict_item(v, k, context) for k, v in example.items()})
    if isinstance(example, t.Sequence) and not isinstance(example, str):
        if isinstance(example, (tuple, list)):
            _type = type(example)
        else:
            _type = None
        return randlist(
            type=_type,
            rnd=context.rnd,
            items_list=tuple(_list_item(exm, context) for exm in enumerate(example)),
        )
    if isinstance(example, t.Iterator):
        return by_iterator(example)
    if isinstance(example, t.Callable):
        return by_callable(example)

    return context.from_example(type(example))



import enum
import typing as t


class Linesep(enum.Enum):
    LF = "\n"
    CRLF = "\r\n"
    CR = "\r"

    @classmethod
    def names(cls) -> t.Tuple[str, ...]:
        return tuple(map(lambda e: e.name, cls))



import datetime as dt
import typing as t
from random import Random

from ._base import Factory
from .._utils.nullsafe import dfor
from ..exceptions import FactoryConstructionError


def randdate(
    minimum: t.Union[dt.date, dt.datetime, None] = None,
    maximum: t.Union[dt.date, dt.datetime, None] = None,
    *,
    rnd: t.Optional[Random] = None,
) -> Factory[dt.date]:
    """Return a factory generating random date values.

    Parameters
    ----------
    minimum : date | datetime, optional
        the minimum
    maximum : date | datetime, optional
        the maximum
    rnd : Random, optional
        random number generator to be used

    Raises
    ------
    FactoryConstructionError
        When the specified generating conditions are inconsistent.
    """
    return DateRandomFactory(
        minimum,
        maximum,
        rnd=rnd,
    )


class DateRandomFactory(Factory[dt.date]):
    """factory generating random date values"""

    _random: Random
    _min: dt.date
    _max: dt.date
    _range: int

    def __init__(
        self,
        minimum: t.Union[dt.date, dt.datetime, None] = None,
        maximum: t.Union[dt.date, dt.datetime, None] = None,
        *,
        rnd: t.Optional[Random] = None,
    ):
        """Return a factory generating random date values.

        Parameters
        ----------
        minimum : date | datetime, optional
            the minimum
        maximum : date | datetime, optional
            the maximum
        rnd : Random, optional
            random number generator to be used

        Raises
        ------
        FactoryConstructionError
            When the specified generating conditions are inconsistent.
        """
        self._random = dfor(rnd, Random())
        self._min, self._max = self._normalize(minimum, maximum)

        if self._min > self._max:
            raise FactoryConstructionError("empty range for randdate")

        self._range = (self._max - self._min).days

    def _next(self) -> dt.date:
        relative_days = self._random.randint(0, self._range)
        return self._min + dt.timedelta(days=relative_days)

    @classmethod
    def _normalize(
        cls,
        minimum: t.Union[dt.date, dt.datetime, None] = None,
        maximum: t.Union[dt.date, dt.datetime, None] = None,
    ) -> t.Tuple[dt.date, dt.date]:
        if isinstance(minimum, dt.datetime):
            minimum = minimum.date()
        if isinstance(maximum, dt.datetime):
            maximum = maximum.date()

        if minimum is not None and maximum is not None:
            return minimum, maximum
        elif minimum is not None:
            return minimum, minimum + dt.timedelta(days=364)
        elif maximum is not None:
            return maximum - dt.timedelta(days=364), maximum
        else:
            now = dt.datetime.utcnow().date()
            return now - dt.timedelta(days=182), now + dt.timedelta(hours=182)



from random import Random
import typing as t

from ._base import Factory
from ._choice import randchoice


def const(
    value: t.Any,
    rnd: t.Optional[Random] = None,
) -> Factory[t.Any]:
    """Return a factory which returns the specified value.

    Parameters
    ----------
    value : Any
        the value
    rnd : Random, optional
        It is not normally used, but it can be accepted as an argument
        to match other Factory construction functions.
    """
    return randchoice(value, rnd=rnd)



import typing as t

from .. import Subcmd
from ._base import SubcmdDef
from ._bool import SubcmdDefBool
from ._byfile import SubcmdDefByfile
from ._date import SubcmdDefDate
from ._datetime import SubcmdDefDatetime
from ._decimal import SubcmdDefDecimal
from ._float import SubcmdDefFloat
from ._int import SubcmdDefInt
from ._str import SubcmdDefString
from ._time import SubcmdDefTime
from ._timedelta import SubcmdDefTimedelta

_subcmd_def_list: t.Sequence[SubcmdDef] = (
    SubcmdDefBool(),
    SubcmdDefByfile(),
    SubcmdDefDate(),
    SubcmdDefDatetime(),
    SubcmdDefDecimal(),
    SubcmdDefFloat(),
    SubcmdDefInt(),
    SubcmdDefString(),
    SubcmdDefTime(),
    SubcmdDefTimedelta(),
)

_subcmd_def_dict: t.Mapping[Subcmd, SubcmdDef] = {
    subcmd_def.cmd(): subcmd_def for subcmd_def in _subcmd_def_list
}


def iter_subcmd_def() -> t.Iterator[SubcmdDef]:
    return iter(_subcmd_def_list)


def get_subcmd_def(subcmd: Subcmd) -> SubcmdDef:
    return _subcmd_def_dict[subcmd]



import datetime as dt
import typing as t
from random import Random

from ._base import Factory
from .._utils.nullsafe import dfor
from ..exceptions import FactoryConstructionError
from ._datetime import DatetimeRandomFactory


def randtime(
    minimum: t.Optional[dt.time] = None,
    maximum: t.Optional[dt.time] = None,
    *,
    tzinfo: t.Union[t.Literal[False], None, dt.tzinfo] = False,
    rnd: t.Optional[Random] = None,
) -> Factory[dt.time]:
    """Return a factory generating random time values.

    Parameters
    ----------
    minimum : time, optional
        the minimum
    maximum : time, optional
        the maximum
    tzinfo : tzinfo | None, optional
        If specified, the tzinfo of result will be fixed to this value
        (False means no specification).
        When it fixes aware datetime to aware, the time is corrected.
        Otherwise, only the tzinfo is changed.
    rnd : Random, optional
        random number generator to be used

    Raises
    ------
    FactoryConstructionError
        When the specified generating conditions are inconsistent.
    """
    return TimeRandomFactory(
        minimum,
        maximum,
        fix_timezone=tzinfo is not False,
        fixed_timezone=tzinfo if tzinfo is not False else None,
        rnd=rnd,
    )


class TimeRandomFactory(Factory[dt.time]):
    """factory generating random time values"""

    _random: Random
    _min: dt.time
    _max: dt.time
    _base: Factory[dt.time]

    def __init__(
        self,
        minimum: t.Union[dt.time, None] = None,
        maximum: t.Union[dt.time, None] = None,
        *,
        fix_timezone: bool = False,
        fixed_timezone: t.Optional[dt.tzinfo] = None,
        rnd: t.Optional[Random] = None,
    ):
        """Return a factory generating random time values.

        Parameters
        ----------
        minimum : time, optional
            the minimum
        maximum : time, optional
            the maximum
        fix_timezone : bool, default=False
            If it is True, the tzinfo of result is fixed to `fixed_timezone`.
        fixed_timezone : tzinfo, default=None
            If `fix_timezone` is True, the tzinfo of result is fixed to this.
        rnd : Random, optional
            random number generator to be used
        """
        self._random = dfor(rnd, Random())
        self._min, self._max = self._normalize(minimum, maximum)
        self._fix_timezone = fix_timezone
        self._fixed_timezone = fixed_timezone

        if (self._min.tzinfo is None and self._max.tzinfo is not None) or (
            self._min.tzinfo is not None and self._max.tzinfo is None
        ):
            raise FactoryConstructionError(
                "cannot define range for randtime with a naive time and an aware time"
            )

        if self._max >= self._min:
            dummy_date = dt.date(1970, 7, 1)
            min_datetime = dt.datetime.combine(dummy_date, self._min)
            max_datetime = dt.datetime.combine(dummy_date, self._max)
        else:
            dummy_date1 = dt.date(1970, 7, 1)
            dummy_date2 = dt.date(1970, 7, 2)
            min_datetime = dt.datetime.combine(dummy_date1, self._min)
            max_datetime = dt.datetime.combine(dummy_date2, self._max)

        self._base = DatetimeRandomFactory(
            min_datetime,
            max_datetime,
            fix_timezone=fix_timezone,
            fixed_timezone=fixed_timezone,
            rnd=rnd,
        ).post_process(lambda v: v.time().replace(tzinfo=v.tzinfo))

    def _next(self) -> dt.time:
        return self._base.next()

    @classmethod
    def _normalize(
        cls,
        minimum: t.Union[dt.time, None],
        maximum: t.Union[dt.time, None],
    ) -> t.Tuple[dt.time, dt.time]:
        if minimum is not None and maximum is not None:
            return minimum, maximum
        elif minimum is not None:
            return minimum, minimum.replace(hour=(minimum.hour + 1) % 24)
        elif maximum is not None:
            return maximum.replace(hour=(maximum.hour + 23) % 24), maximum
        else:
            return dt.time(0, 0, 0), dt.time(23, 59, 59, 999_999)



"""
the package contains the Args of module execution and its builder
"""

import argparse
import codecs
import datetime
import itertools
import typing as t

from . import Linesep
from ._subcmd import Subcmd


class Args:
    """Argument object when using randog as command"""

    _args: argparse.Namespace
    _commanded_args: t.Sequence[str]

    def __init__(self, argv: t.Sequence[str]):
        from ._subcmd_def import iter_subcmd_def

        self._commanded_args = argv[1:]

        parser = argparse.ArgumentParser(
            prog="randog",
            description="It generates values randomly according to the specified mode "
            "and arguments.",
        )
        subparsers = parser.add_subparsers(
            required=True,
            dest="sub",
            metavar="MODE",
            help=(
                "mode of value generation; candidates: "
                f"{', '.join(map(lambda c: c.value, Subcmd))}. "
                "For more information, see the command 'randog MODE --help'."
            ),
        )

        # create subcommands and add them to the ArgumentParser
        subcmd_parsers = {}
        for subcmd in iter_subcmd_def():
            subcmd_parsers[subcmd.cmd()] = subcmd.add_parser(subparsers)

        # parse the arguments
        self._args = parser.parse_args(self.commanded_args)

        # validate arguments for subcommands
        for subcmd in iter_subcmd_def():
            subcmd.validate_parser(self, subcmd_parsers[subcmd.cmd()])

    @property
    def commanded_args(self) -> t.Sequence[str]:
        return self._commanded_args

    @property
    def sub_cmd(self) -> Subcmd:
        return Subcmd(self._args.sub)

    @property
    def factories(self) -> t.Sequence[str]:
        return self._args.factories

    @property
    def repeat(self) -> int:
        return self._args.repeat

    @property
    def list(self) -> t.Optional[int]:
        return self._args.list

    @property
    def output_fmt(self) -> t.Optional[str]:
        return self._args.output_fmt

    @property
    def output_path(self) -> t.Optional[str]:
        return self._args.output

    @property
    def output_encoding(self) -> t.Optional[str]:
        specified = self._args.output_encoding
        if specified is None or self.output_path is None:
            return None
        try:
            codecs.lookup(specified)
            return specified
        except LookupError:
            raise ValueError(f"illegal encoding: {specified}")

    @property
    def output_linesep(self) -> t.Optional[str]:
        specified = self._args.output_linesep
        if specified is None or self.output_path is None:
            return None
        try:
            return Linesep[specified].value
        except KeyError:
            raise ValueError(f"illegal linesep: {specified}")

    @property
    def iso(self) -> bool:
        if hasattr(self._args, "iso"):
            return self._args.iso
        else:
            return False

    @property
    def format(self) -> t.Optional[str]:
        if hasattr(self._args, "format"):
            return self._args.format
        else:
            return None

    @property
    def regenerate(self) -> float:
        if hasattr(self._args, "regenerate"):
            return self._args.regenerate
        else:
            return 0.0

    @property
    def discard(self) -> float:
        if hasattr(self._args, "discard"):
            return self._args.discard
        else:
            return 0.0

    @property
    def csv(self) -> t.Optional[int]:
        if hasattr(self._args, "csv"):
            return self._args.csv
        else:
            return None

    @property
    def error_on_factory_stopped(self) -> bool:
        return self.get("error_on_factory_stopped", False)

    @property
    def hide_randog_warnings(self) -> bool:
        return self.get("quiet", False)

    @property
    def log_stderr(self) -> t.Optional[str]:
        return self.get("log_stderr", None)

    @property
    def log_config_file(self) -> t.Optional[str]:
        return self.get("log", None)

    @property
    def env(self) -> t.Mapping[str, str]:
        if self._args.env is None:
            return {}

        result = {}
        for item in itertools.chain(*self._args.env):
            key_value = item.split("=", 1)
            if len(key_value) <= 1:
                result[key_value[0]] = ""
            else:
                key, value = key_value
                result[key] = value
        return result

    def output_path_for(
        self,
        number: int,
        *,
        def_file: str,
        repeat_count: int,
        factory_count: int,
        now: datetime.datetime,
        env: t.Mapping[str, str],
    ) -> t.Optional[str]:
        if self._args.output is None:
            return None
        else:
            return self._args.output.format(
                number,
                def_file=def_file,
                repeat_count=repeat_count,
                factory_count=factory_count,
                now=now,
                **env,
            )

    def get(self, key: str, default: t.Any = None) -> t.Any:
        if hasattr(self._args, key):
            return getattr(self._args, key)
        else:
            return default



from setuptools import setup, find_packages

import randog


def load_readme() -> str:
    with open("README.md", encoding="utf-8") as f:
        return f.read()


setup(
    name="randog",
    version=randog.__version__,
    project_urls={
        "Bug Tracker": "https://github.com/unaguna/random-obj-generator/issues",
        "Documentation": "https://unaguna.github.io/random-obj-generator/",
        "Source Code": "https://github.com/unaguna/random-obj-generator",
    },
    author="k-izumi",
    author_email="k.izumi.ysk@gmail.com",
    maintainer="k-izumi",
    maintainer_email="k.izumi.ysk@gmail.com",
    description="Generate data randomly",
    long_description=load_readme(),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=("tests",)),
    install_requires=[],
    license="MIT",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Topic :: Utilities",
    ],
)



from random import Random
import typing as t

from ._base import Factory
from .._utils.nullsafe import dfor
from ..exceptions import FactoryConstructionError


def randint(
    minimum: int,
    maximum: int,
    *,
    rnd: t.Optional[Random] = None,
) -> Factory[int]:
    """Return a factory generating random int values.

    Parameters
    ----------
    minimum : int
        the minimum
    maximum : int
        the maximum
    rnd : Random, optional
        random number generator to be used

    Raises
    ------
    FactoryConstructionError
        When the specified generating conditions are inconsistent.
    """
    return IntRandomFactory(minimum, maximum, rnd=rnd)


class IntRandomFactory(Factory[int]):
    """factory generating random int values"""

    _random: Random
    _min: int
    _max: int

    def __init__(
        self,
        minimum: int,
        maximum: int,
        *,
        rnd: t.Optional[Random] = None,
    ):
        """Return a factory generating random int values.

        Parameters
        ----------
        minimum : int
            the minimum
        maximum : int
            the maximum
        rnd : Random, optional
            random number generator to be used

        Raises
        ------
        FactoryConstructionError
            When the specified generating conditions are inconsistent.
        """
        self._random = dfor(rnd, Random())
        self._min = minimum
        self._max = maximum

        if minimum > maximum:
            raise FactoryConstructionError("empty range for randint")

    def _next(self) -> int:
        return self._random.randint(self._min, self._max)



from random import Random
import typing as t

from ._base import Factory
from .._utils.nullsafe import dfor
from ..exceptions import FactoryConstructionError


def randfloat(
    minimum: t.Optional[t.SupportsFloat] = None,
    maximum: t.Optional[t.SupportsFloat] = None,
    *,
    p_inf: t.SupportsFloat = 0.0,
    n_inf: t.SupportsFloat = 0.0,
    nan: t.SupportsFloat = 0.0,
    rnd: t.Optional[Random] = None,
) -> Factory[float]:
    """Return a factory generating random float values.

    Parameters
    ----------
    minimum : float, optional
        the minimum
    maximum : float, optional
        the maximum
    p_inf : float, default=0
        the probability of positive infinity
    n_inf : float, default=0
        the probability of negative infinity
    nan : float, default=0
        the probability of NaN
    rnd : Random, optional
        random number generator to be used

    Raises
    ------
    FactoryConstructionError
        When the specified generating conditions are inconsistent.
    """
    return FloatRandomFactory(
        minimum,
        maximum,
        p_inf=p_inf,
        n_inf=n_inf,
        nan=nan,
        rnd=rnd,
    )


class FloatRandomFactory(Factory[float]):
    """factory generating random float values"""

    _random: Random
    _min: float
    _max: float
    _p_inf: float
    _n_inf: float
    _nan: float

    def __init__(
        self,
        minimum: t.Optional[t.SupportsFloat] = None,
        maximum: t.Optional[t.SupportsFloat] = None,
        *,
        p_inf: t.SupportsFloat = 0.0,
        n_inf: t.SupportsFloat = 0.0,
        nan: t.SupportsFloat = 0.0,
        rnd: t.Optional[Random] = None,
    ):
        """Return a factory generating random float values.

        Parameters
        ----------
        minimum : float, optional
            the minimum
        maximum : float, optional
            the maximum
        p_inf : float, default=0
            the probability of positive infinity
        n_inf : float, default=0
            the probability of negative infinity
        nan : float, default=0
            the probability of NaN
        rnd : Random, optional
            random number generator to be used

        Raises
        ------
        FactoryConstructionError
            When the specified generating conditions are inconsistent.
        """
        self._random = dfor(rnd, Random())
        self._min, self._max = self._normalize(minimum, maximum)
        self._p_inf = float(p_inf)
        self._n_inf = float(n_inf)
        self._nan = float(nan)

        if self._min > self._max:
            raise FactoryConstructionError("empty range for randfloat")
        if self._p_inf < 0.0 or self._n_inf < 0.0 or self._nan < 0.0:
            raise FactoryConstructionError(
                "the probabilities `p_inf`, `n_inf`, and `nan` must range from 0 to 1"
            )
        if self._p_inf + self._n_inf + self._nan > 1.0:
            raise FactoryConstructionError(
                "the sum of probabilities `p_inf`, `n_inf`, and `nan` "
                "must range from 0 to 1"
            )

    def _next(self) -> float:
        pre_weight = self._random.random()
        pre_weight -= self._p_inf
        if pre_weight < 0:
            return float("inf")
        pre_weight -= self._n_inf
        if pre_weight < 0:
            return float("-inf")
        pre_weight -= self._nan
        if pre_weight < 0:
            return float("NaN")

        weight = self._random.random()
        return self._max * weight + self._min * (1 - weight)

    @classmethod
    def _normalize(
        cls,
        minimum: t.Optional[t.SupportsFloat],
        maximum: t.Optional[t.SupportsFloat],
    ) -> t.Tuple[float, float]:
        default_range = 1.0

        if minimum is None and maximum is None:
            return 0.0, 1.0
        elif minimum is None:
            maximum = float(maximum)
            return maximum - default_range, maximum
        elif maximum is None:
            minimum = float(minimum)
            return minimum, minimum + default_range
        else:
            return float(minimum), float(maximum)
